from dataclasses import dataclass, field
from types import SimpleNamespace
from typing import List, Tuple, Hashable, Union, Iterable, Set, Dict

import numpy as np
from ordered_set import OrderedSet

from respice.analysis import UniqueEdgeMultiDiGraph
from respice.collections import OneToManyInvertibleMapping
from respice.components import Branch
from respice.components.Branch import VoltageBranch, CurrentBranch

from networkx.algorithms.components import weakly_connected_components


@dataclass
class KirchhoffEquationTerm:
    negative: bool
    branch: CurrentBranch
    coupled_branches: List[Union[VoltageBranch, CurrentBranch]]


@dataclass
class KirchhoffEquationBranchConsecutiveTerm:
    negative: bool
    branch: VoltageBranch


@dataclass
class KirchhoffEquation:
    terms: List[Union[KirchhoffEquationTerm, KirchhoffEquationBranchConsecutiveTerm]] = field(default_factory=list)


@dataclass
class BranchConsecutiveEquation:
    branch: VoltageBranch
    potentials: Tuple[Hashable, Hashable]
    coupled_branches: List[Union[VoltageBranch, CurrentBranch]]


class MNAEquationStack:
    """
    Manages the master equation that comprises an electrical circuit system using the so called MNA - Modified Nodal
    Analysis. Used inside the simulation algorithm to find the final solution.

    The MNA equation stack automatically generates an inner representation of all mathematical terms and equations
    necessary to describe the problem. The problem is a multivariate non-linear root-finding problem.

    This class is majorly to be used via solvers like scipy's fsolve function, where you feed in evaluate() or
    the optimized expression returned by lambdify().

    For more information about the MNA itself, see https://en.wikipedia.org/wiki/Modified_nodal_analysis
    """

    def __init__(self, graph: UniqueEdgeMultiDiGraph, couplings: Iterable[List[Branch]], references: Set[Hashable] = frozenset()):
        """
        Create a new equation system for a circuit.

        :param graph:
            The circuit graph that stores the connectivity between branches.
        :param couplings:
            The coupled branches. Coupled branches will receive the same voltage-current-vector ("coupling-vector") so
            they can describe mutual dynamics between themselves (e.g. for coupled inductors).
        :param references:
            Ground nodes to prefer.
            Invalid or conflicting ground nodes are ignored or resolved automatically by picking the first eligible
            node. If no ground nodes are specified or if ground nodes are missing for a connected subgraph where another
            ground node is required, it is automatically selected.
        """
        self._cache = SimpleNamespace()
        self._cache.equation_lambdified = None
        self._cache.jacobian_lambdified = None

        self._graph: UniqueEdgeMultiDiGraph = graph
        self._reference_nodes = OneToManyInvertibleMapping()

        # Stores the vector-index relations to potential nodes.
        self._nodes: OrderedSet[Hashable] = OrderedSet()
        # Stores the vector-index relations to branch currents when voltage-branches are used. The branch consecutive
        # currents are stored after the nodes in the input vector for evaluate().
        self._currents: OrderedSet[VoltageBranch] = OrderedSet()

        self._kirchhoff_equations: List[KirchhoffEquation] = []
        self._branch_consecutive_equations: List[BranchConsecutiveEquation] = []

        self._coupling_dictionary = {branch: coupled_branches
                                     for coupled_branches in couplings
                                     for branch in coupled_branches}

        for subset in weakly_connected_components(graph):
            # A bug in NetworkX prohibits editing the yielded sets directly, so we need to make a copy.
            # See https://github.com/networkx/networkx/issues/4331.
            subset = set(subset)

            # Pin down ground node. Depending on the given preferred reference nodes, either the immediate next node is
            # picked or one of the references specified. If more than one preferred ground node is specified that belong
            # to the same subset, the first available one is picked.
            preferred_ref_nodes = subset & references
            if preferred_ref_nodes:
                ref_node = preferred_ref_nodes.pop()
                subset.remove(ref_node)
            else:
                ref_node = subset.pop()

            for node in subset:
                self._reference_nodes[ref_node] = node

            subset_iter = iter(subset)

            for potential in subset_iter:
                self._nodes.add(potential)

                kirchhoff_equation = KirchhoffEquation()

                for negative, branches in [(False, graph.in_edges(potential, keys=True)),
                                           (True, graph.out_edges(potential, keys=True))]:
                    for source, target, branch in branches:
                        # Add equations according to MNA (Modified Nodal Analysis) technique.
                        if isinstance(branch, CurrentBranch):
                            term = KirchhoffEquationTerm(
                                negative=negative,
                                branch=branch,
                                coupled_branches=self._coupling_dictionary[branch],
                            )
                        elif isinstance(branch, VoltageBranch):
                            term = KirchhoffEquationBranchConsecutiveTerm(
                                negative=negative,
                                branch=branch,
                            )

                            # Branch consecutive equations are processed in a later step for more efficient iteration
                            # over nodes. Since branches are always connected between two nodes, but we can just have
                            # one branch consecutive equation per branch, we would need to track that information if we
                            # do it here.

                        else:
                            raise ValueError(f'Unknown branch type encountered: {type(branch)}')

                        kirchhoff_equation.terms.append(term)

                self._kirchhoff_equations.append(kirchhoff_equation)

        # Now processing branch consecutive equations.
        for source, target, branch in graph.edges(keys=True):
            if isinstance(branch, VoltageBranch):
                self._branch_consecutive_equations.append(BranchConsecutiveEquation(
                    branch=branch,
                    potentials=(source, target),
                    coupled_branches=self._coupling_dictionary[branch],
                ))

                self._currents.add(branch)

    def __len__(self):
        """
        :return:
            The number of equations inside the equation stack.
        """
        return len(self._nodes) + len(self._currents)

    def _expand_jacobi_vector(self, jacobi, coupled_branches: List[Union[VoltageBranch, CurrentBranch]]) -> np.ndarray:
        # Properly creates a Jacobian vector suiting the problem from a component-tailored Jacobian.
        # Effectively increases dimension and inserts zeroes accordingly so the Jacobian's from different branches can
        # be easily blended together.

        jacobi_vector = np.zeros(len(self))

        for i, branch in enumerate(coupled_branches):
            if isinstance(branch, VoltageBranch):
                jacobi_vector[self._currents.index(branch) + len(self._nodes)] += jacobi[i]
            elif isinstance(branch, CurrentBranch):
                source, target = self._graph.get_nodes(branch)

                # Although the Jacobian is calculated for the voltage difference over the branch, one can easily
                # prove that it can still be used to get the right Jacobian suited for the target problem. Assume
                # that the voltage over a branch v = b - a. b is the target voltage and a the source voltage. Our
                # problem is that we have df/dv at hand, but we need df/da and df/db instead.
                # Using substitution rule one can derive dv = db and dv = -da. Then we get
                # df/dv = df/db and -df/dv = df/da.
                if target in self._nodes:
                    jacobi_vector[self._nodes.index(target)] += jacobi[i]
                if source in self._nodes:
                    jacobi_vector[self._nodes.index(source)] += -jacobi[i]
            else:
                raise AssertionError(f'Unknown branch type encountered: {type(branch)}')

        return jacobi_vector

    def _create_coupled_v_i_vector(self,
                                   v_i: np.ndarray,
                                   coupled_branches: List[Union[VoltageBranch, CurrentBranch]]) -> np.ndarray:
        # Creates the voltage-current-vector ("coupling-vector") for a set of coupled branches from a solution.

        vec = []

        for branch in coupled_branches:
            if isinstance(branch, VoltageBranch):
                val = v_i[self._currents.index(branch) + len(self._nodes)]
            elif isinstance(branch, CurrentBranch):
                source, target = self._graph.get_nodes(branch)
                val = (
                    # If the nodes are not there, they were classified as reference nodes and would have a value of
                    # zero.
                    (v_i[self._nodes.index(target)] if target in self._nodes else 0) -
                    (v_i[self._nodes.index(source)] if source in self._nodes else 0)
                )
            else:
                raise AssertionError(f'Unknown branch type encountered: {type(branch)}')

            vec.append(val)

        return np.array(vec)

    def evaluate(self, v_i: np.ndarray, t1: float, t2: float) -> np.ndarray:
        """
        Evaluates the equation stack.

        :param v_i:
            The voltages (measured in Volts) and branch consecutive currents (measured in Ampere) as input.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The result.
        """
        out_vector = []

        for kirchhoff_equation in self._kirchhoff_equations:
            summation = 0.0

            for term in kirchhoff_equation.terms:
                if isinstance(term, KirchhoffEquationTerm):
                    summation += (-1 if term.negative else 1) * term.branch.get_current(
                        self._create_coupled_v_i_vector(v_i, term.coupled_branches),
                        t1,
                        t2,
                    )
                elif isinstance(term, KirchhoffEquationBranchConsecutiveTerm):
                    summation += (
                        (-1 if term.negative else 1) *
                        v_i[self._currents.index(term.branch) + len(self._nodes)]
                    )
                else:
                    raise AssertionError(f'Invalid term type encountered: {type(term)}')

            out_vector.append(summation)

        for branch_consecutive_equation in self._branch_consecutive_equations:
            source, target = branch_consecutive_equation.potentials

            out_vector.append(
                branch_consecutive_equation.branch.get_voltage(
                    self._create_coupled_v_i_vector(v_i, branch_consecutive_equation.coupled_branches),
                    t1,
                    t2,
                )
                + (v_i[self._nodes.index(source)] if source in self._nodes else 0)
                - (v_i[self._nodes.index(target)] if target in self._nodes else 0)
            )

        return np.array(out_vector)

    def get_voltage_between_nodes(self, solution: np.ndarray, source, target) -> float:
        """
        Returns the voltage between two nodes from a calculated solution.

        :param solution:
            The solution vector.
        :param source:
            The source node.
        :param target:
            The target node.
        :return:
            The voltage between `source` and `target` nodes.
        """
        return (
            (solution[self._nodes.index(target)] if target in self._nodes else 0) -
            (solution[self._nodes.index(source)] if source in self._nodes else 0)
        )

    def get_reference_nodes(self):
        """
        Return all ground nodes (a.k.a. GND or reference nodes).
        """
        return self._reference_nodes.keys()

    def get_node_voltage(self, solution: np.ndarray, node: Hashable) -> float:
        """
        Returns the voltage (or potential) of a node from a given solution.

        :param solution:
            The solution to get the voltage from.
        :param node:
            The node.
        :return:
            The voltage/potential measured in Volts.
        """
        return solution[self._nodes.index(node)]

    def get_node_voltages(self, solution: np.ndarray) -> Dict[Hashable, float]:
        """
        Returns the voltages (or potentials) from all nodes.

        :param solution:
            The solution to get the voltages from.
        :return:
            A dictionary mapping nodes to their respective voltage/potential.
        """
        vs = {node: self.get_node_voltage(solution, node) for node in self._nodes}
        vs.update({gnd_node: 0 for gnd_node in self._reference_nodes})
        return vs

    def get_voltage(self, solution: np.ndarray, branch: CurrentBranch) -> float:
        """
        Returns the voltage for a given current-branch from a calculated solution.

        Since only this stack knows how branches and their voltages and currents are tracked in the solution vector,
        use this function afterwards to reassign those values back to the actual branches.

        :param solution:
            The solution vector.
        :param branch:
            The current-branch to get the voltage for.
        :return:
            The voltage for `branch`.
        """
        source, target = self._graph.get_nodes(branch)
        return self.get_voltage_between_nodes(solution, source, target)

    def get_current(self, solution: np.ndarray, branch: VoltageBranch) -> float:
        """
        Returns the current for a given voltage-branch from a calculated solution.

        Since only this stack knows how branches and their voltages and currents are tracked in the solution vector,
        use this function afterwards to reassign those values back to the actual branches.

        :param solution:
            The solution vector.
        :param branch:
            The voltage-branch to get the current for.
        :return:
            The current for `branch`.
        """
        return solution[self._currents.index(branch) + len(self._nodes)]

    def assemble_vector(self, v: Dict[Hashable, float], i: Dict[VoltageBranch, float], fill: float = 0.0):
        """
        Assembles a proper vector from voltages and currents that can be used to evaluate this equation stack.

        :param v:
            Mapping of nodes to voltages.
        :param i:
            Mapping of voltage branches to their respective currents (other branch types are insignificant).
        :param fill:
            Fill value to use if a needed node or branch is missing.
        :return:
            A vector that can be used for `evaluate` and `jacobian` for equation evaluation.
        """
        vec = np.full(len(self), fill, dtype=float)
        for node in v:
            if node in self._nodes:
                # Subtract also the reference potential if one was given explicitly to properly offset all voltages
                # within a joint group of nodes.
                vec[self._nodes.index(node)] = v[node] - v.get(self._reference_nodes.reverse[node], 0)

        for branch in i:
            if branch in self._currents:
                vec[self._currents.index(branch) + len(self._nodes)] = i[branch]

        return vec

    def disassemble_vector(self, solution: np.ndarray):
        """
        Reverse operation of `assemble_vector`, creates `node -> potential` and `branch -> current` dictionaries
        from a solution.

        :param solution:
            The solution to create the dictionaries from.
        :return:
            `node -> potential` and `branch -> current` dicts as a 2-tuple.
        """
        v = {}
        i = {}

        for branch in self._currents:
            i[branch] = self.get_current(solution, branch)

        for node in self._nodes:
            v[node] = self.get_node_voltage(solution, node)

        for gnd_node in self._reference_nodes:
            v[gnd_node] = 0

        return v, i

    def _create_lambdified_coupled_v_i_vector(self, coupled_branches: List[Union[VoltageBranch, CurrentBranch]]) -> List[str]:
        coupled_entries = []

        for branch in coupled_branches:
            if isinstance(branch, VoltageBranch):
                coupled_val = f'v_i[{self._currents.index(branch) + len(self._nodes)}]'
            elif isinstance(branch, CurrentBranch):
                source, target = self._graph.get_nodes(branch)

                coupled_val = ''
                if source is target:
                    # Coupling that self-closes on a ground node might end up here.
                    coupled_val = '0'
                else:
                    if target in self._nodes:
                        coupled_val += f'v_i[{self._nodes.index(target)}]'
                    if source in self._nodes:
                        coupled_val += f'-v_i[{self._nodes.index(source)}]'
            else:
                raise AssertionError(f'Unknown branch type encountered: {type(branch)}')

            coupled_entries.append(coupled_val)

        return coupled_entries

    def lambdify(self):
        """
        Optimizes the MNA equation stack by returning a lambdified expression.

        This function compiles a Python function that immediately computes `evaluate()`.

        Note that the returned object is immediately callable and does not have an `evaluate()` member function.
        Also later updates on the instance of the MNA equation stack will not reflect into already lambdified
        expressions. In this case it has to be recompiled.

        :return:
            A function that is callable like `evaluate()` and behaves exactly like it.
        """
        if self._cache.equation_lambdified is not None:
            return self._cache.equation_lambdified

        mapped_branches = OrderedSet()
        code_globals = {'branches': mapped_branches}

        vector_entries = []

        for equation in self._kirchhoff_equations:
            terms = []

            for term in equation.terms:
                if isinstance(term, KirchhoffEquationTerm):
                    val = (f'branches[{mapped_branches.add(term.branch)}]'
                           f'.get_current(np.array(['
                           f'{",".join(self._create_lambdified_coupled_v_i_vector(term.coupled_branches))}'
                           f']),t1,t2)')

                elif isinstance(term, KirchhoffEquationBranchConsecutiveTerm):
                    val = f'v_i[{self._currents.index(term.branch) + len(self._nodes)}]'
                else:
                    raise AssertionError(f'Unexpected equation term type: {type(term)}')

                if term.negative:
                    val = '-' + val

                terms.append(val)

            vector_entries.append('+'.join(terms))

        for equation in self._branch_consecutive_equations:
            source, target = equation.potentials
            vector_entries.append(
                f'branches[{mapped_branches.add(equation.branch)}]'
                f'.get_voltage(np.array(['
                f'{",".join(self._create_lambdified_coupled_v_i_vector(equation.coupled_branches))}'
                f']),t1,t2)'
                f'{f"+v_i[{self._nodes.index(source)}]" if source in self._nodes else ""}'
                f'{f"-v_i[{self._nodes.index(target)}]" if target in self._nodes else ""}'
            )

        code = [
            'import numpy as np',
            'def eq(v_i: np.ndarray, t1: float, t2: float) -> np.ndarray:',
            ' return np.array([',
        ] + [entry + ',' for entry in vector_entries] + [
            ' ])',
        ]

        # Compile the expression and get back the set up eq() function.
        code_object = compile('\n'.join(code), '<lambdified-mna-equation-stack>', 'exec')
        exec(code_object, code_globals)
        fun = code_globals['eq']

        self._cache.equation_lambdified = fun

        return fun

    def jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> np.ndarray:
        """
        Evaluates the Jacobian of the equation stack.

        :param v_i:
            The voltages (measured in Volts) and branch consecutive currents (measured in Ampere) as input.
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The result.
        """
        jacobi_matrix = []

        for kirchhoff_equation in self._kirchhoff_equations:
            jacobi_vector = np.zeros(len(self))

            for term in kirchhoff_equation.terms:
                if isinstance(term, KirchhoffEquationTerm):
                    jacobi_vector += (-1 if term.negative else 1) * self._expand_jacobi_vector(
                        term.branch.get_jacobian(
                            self._create_coupled_v_i_vector(v_i, term.coupled_branches),
                            t1,
                            t2,
                        ),
                        term.coupled_branches,  # TODO Maybe I should move all this into a function, since I reuse coupled_branches twice.
                    )
                elif isinstance(term, KirchhoffEquationBranchConsecutiveTerm):
                    jacobi_vector[self._currents.index(term.branch) + len(self._nodes)] += 1.0
                else:
                    raise AssertionError(f'Invalid term type encountered: {type(term)}')

            jacobi_matrix.append(jacobi_vector)

        for branch_consecutive_equation in self._branch_consecutive_equations:
            jacobi_vector = np.zeros(len(self))

            source, target = branch_consecutive_equation.potentials

            jacobi_vector += self._expand_jacobi_vector(
                branch_consecutive_equation.branch.get_jacobian(
                    self._create_coupled_v_i_vector(v_i, branch_consecutive_equation.coupled_branches),
                    t1,
                    t2,
                ),
                branch_consecutive_equation.coupled_branches,  # TODO Maybe I should move all this into a function, since I reuse coupled_branches twice.
            )

            # We have 2 additional entries for branch consecutive equations: 1 and -1
            # respectively for source and target voltage Jacobians.
            if source in self._nodes:
                jacobi_vector[self._nodes.index(source)] = 1.0
            if target in self._nodes:
                jacobi_vector[self._nodes.index(target)] = -1.0

            jacobi_matrix.append(jacobi_vector)

        return np.array(jacobi_matrix)

    def lambdify_jacobian(self):
        """
        Creates an optimized version of `jacobian` by compiling a more efficient Python expression.

        :return:
            A callable object that behaves like `jacobian` but faster.
        """
        if self._cache.jacobian_lambdified is not None:
            return self._cache.jacobian_lambdified

        mapped_branches = OrderedSet()
        code_globals = {'branches': mapped_branches}

        jacobi_matrix = []

        for kirchhoff_equation in self._kirchhoff_equations:
            jacobi_vector = [[] for _ in range(len(self))]

            for term in kirchhoff_equation.terms:
                if isinstance(term, KirchhoffEquationTerm):
                    jacobi = f'{"-" if term.negative else ""}jacobians[{mapped_branches.add(term.branch)}]'

                    for i, branch in enumerate(term.coupled_branches):
                        jacobi_value = f'{jacobi}[{i}]'
                        if isinstance(branch, VoltageBranch):
                            jacobi_vector[self._currents.index(branch) + len(self._nodes)].append(jacobi_value)
                        elif isinstance(branch, CurrentBranch):
                            source, target = self._graph.get_nodes(branch)
                            if target in self._nodes:
                                jacobi_vector[self._nodes.index(target)].append(jacobi_value)
                            if source in self._nodes:
                                jacobi_vector[self._nodes.index(source)].append('-' + jacobi_value)
                        else:
                            raise AssertionError(f'Unknown branch type encountered: {type(branch)}')
                elif isinstance(term, KirchhoffEquationBranchConsecutiveTerm):
                    jacobi_vector[self._currents.index(term.branch) + len(self._nodes)].append('-1' if term.negative else '1')
                else:
                    raise AssertionError(f'Invalid term type encountered: {type(term)}')

            jacobi_matrix.append(jacobi_vector)

        for branch_consecutive_equation in self._branch_consecutive_equations:
            jacobi_vector = [[] for _ in range(len(self))]

            source, target = branch_consecutive_equation.potentials

            jacobi = f'jacobians[{mapped_branches.add(branch_consecutive_equation.branch)}]'

            for i, branch in enumerate(branch_consecutive_equation.coupled_branches):
                jacobi_value = f'{jacobi}[{i}]'
                if isinstance(branch, VoltageBranch):
                    jacobi_vector[self._currents.index(branch) + len(self._nodes)].append(jacobi_value)
                elif isinstance(branch, CurrentBranch):
                    source, target = self._graph.get_nodes(branch)
                    if target in self._nodes:
                        jacobi_vector[self._nodes.index(target)].append(jacobi_value)
                    if source in self._nodes:
                        jacobi_vector[self._nodes.index(source)].append('-' + jacobi_value)
                else:
                    raise AssertionError(f'Unknown branch type encountered: {type(branch)}')

            if source in self._nodes:
                jacobi_vector[self._nodes.index(source)].append('1')
            if target in self._nodes:
                jacobi_vector[self._nodes.index(target)].append('-1')

            jacobi_matrix.append(jacobi_vector)

        precalculated_jacobians = [
            (f'branches[{i}].get_jacobian(np.array(['
             f'{",".join(self._create_lambdified_coupled_v_i_vector(self._coupling_dictionary[branch]))}'
             f']),t1,t2)')
            for i, branch in enumerate(mapped_branches)
        ]

        jacobi_matrix_rows = [
            f'[{",".join("+".join(column if column else ["0"]) for column in row)}]'
            for row in jacobi_matrix
        ]

        code = [
            f'import numpy as np',
            f'def jac(v_i: np.ndarray, t1: float, t2: float) -> np.ndarray:',
            f' jacobians = [{",".join(precalculated_jacobians)}]',
            f' return np.array([{",".join(jacobi_matrix_rows)}])',
        ]

        # Compile the expression and get back the set up eq() function.
        code_object = compile('\n'.join(code), '<lambdified-mna-equation-stack-jacobian>', 'exec')
        exec(code_object, code_globals)
        fun = code_globals['jac']

        self._cache.jacobian_lambdified = fun

        return fun



from dataclasses import dataclass
from math import sin, sqrt
from typing import Optional

from .PeriodicCurrentSource import PeriodicCurrentSource


@dataclass(eq=False)
class CurrentSourceHalfWaveAC(PeriodicCurrentSource):
    """
    A half wave / rectified AC current supply.

    You might alternatively set the current amplitude by specifying an effective current.
    Be sure to assign to the property `effective_amplitude` after initialization.

    amplitude:
        The current amplitude of the sine emitted by the supply (measured in Amperes).
    """
    amplitude: float = 1.0

    @property
    def effective_amplitude(self):
        r"""
        The effective amplitude (:math:`A_{RMS}`) is easily derived
        according to the formula for RMS (see https://en.wikipedia.org/wiki/Root_mean_square and
        https://en.wikipedia.org/wiki/RMS_amplitude) and is equal to the normal sine-wave case:

        .. math::

             A_{RMS} &= \sqrt{\frac{1}{T} \int_{0}^{T}{f^2(t) dt}} \\
                     &= A \cdot
                        \sqrt{
                            \frac{1}{2 \pi}
                                \left (
                                    \int_{0}^{T}{|\sin{(t)}|^2 dt}
                                \right )
                            } \\
                     &= A / \sqrt{2}

        :return:
            The effective current (measured in Amperes).
        """
        return self.amplitude / sqrt(2)

    @effective_amplitude.setter
    def effective_amplitude(self, value: float):
        """
        Sets the amplitude of this supply by means of an effective value (RMS).

        This function just multiplies the given value with sqrt(2).

        .. math::

             A = \sqrt{2} \cdot A_{RMS}

        :param value:
            The effective value.
        """
        self.amplitude = value * sqrt(2)

    def get_signal(self, phase: float) -> float:
        return self.amplitude * sin(0.5 * phase)

    def next_signal_event(self, phase1: float, phase2: float) -> Optional[float]:
        if phase1 == 0.0:
            return 0.0



from bisect import bisect_left
from dataclasses import dataclass
from math import pi
from typing import Optional, Iterable

import numpy as np

from .PeriodicSwitch import PeriodicSwitch


@dataclass(eq=False)
class SwitchPWM(PeriodicSwitch):
    """
    A periodic, multi-level PWM modulation source.

    You can supply this source a constant pattern that is repeated throughout a simulation.

    Changing patterns on the same source object is currently not possible. Please instantiate another object with
    the new pattern.
    """
    def __init__(self, pattern_timings: Iterable[float], initial_state: bool = False, *args, **kwargs):
        r"""
        :param pattern_timings:
            The timings between each amplitude change in `pattern_levels` in rad. The range for pattern timings lies
            inside :math:`(0, 2\pi)`. Timings outside this range will be ignored.
        :param initial_state:
            Whether the first state until the next specified pattern timing will be on (`True`) or off (`False`).
        """
        super().__init__(*args, **kwargs)

        self._pattern_timings = list(pattern_timings)
        self._initial_state = initial_state

    def __repr__(self):
        signal_pattern = ''.join(
            '\u2588' if self.get_switch_signal(phase) else '\u2581'
            for phase in np.linspace(0, 2 * pi, 50))

        return f'{type(self).__name__}(pattern: {signal_pattern})'

    def get_switch_signal(self, phase: float) -> bool:
        even = bisect_left(self._pattern_timings, phase) % 2 == 0
        return (even and self._initial_state) or not (even or self._initial_state)

    def next_switch_event(self, phase1: float, phase2: float) -> Optional[float]:
        if phase1 == 0.0 and len(self._pattern_timings) % 2 == 1:
            return 0.0

        a = bisect_left(self._pattern_timings, phase1)
        b = bisect_left(self._pattern_timings, phase2, a)

        if a == b:
            return None

        return self._pattern_timings[a]



from dataclasses import dataclass
from math import exp, log
from types import SimpleNamespace

from respice.math.linalg import TaylorPolynomial
from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent


@dataclass(eq=False)
class LinearizedShockleyDiode(TwoTerminalCurrentComponent):
    """
    Describes an ideal Shockley Diode, linearized after a certain voltage `V_L`.

    Additionally to https://en.wikipedia.org/wiki/Diode_modelling#Shockley_diode_model, for better convergence this
    linearized diode model exists. Exponential terms grow too big too fast and often prevent convergence of the
    solution. A very accurate approximation can be made by linearizing or Taylor-expanding the equations after a certain
    linearization threshold voltage where the curve only grows linearly. So if you encounter convergence problems
    because you have used diodes that are caused by overflow errors or bad convergence properties, this linearized model
    might want you want to use.

    For the pure ideal model, see `respice.components.ShockleyDiode`.

    i_s:
        Reverse bias saturation current (also known as leakage current).
        Depending on the material used, common values are
        * silicon diodes: usually between 10^-12 A and 10^-10 A (1uA to 100uA).
        * germanium diodes: about 10^-4 A (0.1mA)
    v_t:
        Thermal voltage V_T. For approximate room temperate 300K this value
        is 25.85mV (0.02585V). This is default.
    n:
        Ideality factor (also known as quality factor). By default 1.
    dv_th:
        The threshold gradient at which the voltage is linearized / Taylor-expanded (a.k.a. differential resistance).
    order:
        Order :math:`k` of Taylor-expansion to be used behind the threshold voltage.
    """
    i_s: float = 1e-12
    v_t: float = 0.02585
    n: float = 1.0
    dv_th: float = 10000.0
    order: int = 2

    def __post_init__(self):
        super().__post_init__()

        self._cache = SimpleNamespace()
        self._cache.taylor_expansion = SimpleNamespace()
        self._cache.taylor_expansion.key = None
        self._cache.taylor_expansion.value = None

    @property
    def v_th(self):
        return log(self.dv_th * self.v_t / self.i_s) * self.v_t

    def get_ideal_current(self, v: float):
        return self.i_s * (exp(v / (self.n * self.v_t)) - 1)

    def get_ideal_current_derivative(self, v: float, d: int = 1):
        return self.i_s / (self.n * self.v_t)**d * exp(v / (self.n * self.v_t))

    def _get_taylor_expansion(self):
        cachekey = (self.i_s, self.v_t, self.n, self.dv_th, self.order)
        if cachekey is self._cache.taylor_expansion.key:
            return self._cache.taylor_expansion.value

        v_th = self.v_th
        taylor = TaylorPolynomial(
            (
                [self.get_ideal_current(v_th)] +
                [self.get_ideal_current_derivative(v_th, d=i + 1) for i in range(self.order)]
            ),
            v_th,
        )

        self._cache.taylor_expansion.value = taylor
        self._cache.taylor_expansion.key = cachekey

        return taylor

    def get_current(self, v: float, t1: float, t2: float) -> float:
        return (self.get_ideal_current(v)
                if v <= self.v_th else
                self._get_taylor_expansion()(v))

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        return (self.get_ideal_current_derivative(v)
                if v <= self.v_th else
                self._get_taylor_expansion().derivate()(v))



from dataclasses import dataclass, field

import numpy as np

from .TwoTerminalVoltageComponent import TwoTerminalVoltageComponent


@dataclass(eq=False)
class Capacitance(TwoTerminalVoltageComponent):
    """
    Represents an ideal capacitance.

    value:
        The capacitance value (measured in Farad).
    state_voltage:
        The momentary voltage. This is a state variable updated consecutively
        on each simulation iteration. However, you might set it ahead simulation
        to describe the capacitance's initial state.
    state_current:
        The momentary current. Since this component uses the midpoint-method for
        calculation, this additional state variable is introduced for better accuracy.
        This is a state variable updated consecutively on each simulation iteration.
        However, you might set it ahead simulation to describe the capacitance's initial state.
    """
    value: float

    state_voltage: float = field(default=0.0, repr=False)
    state_current: float = field(default=0.0, repr=False)

    def get_voltage(self, i: float, t1: float, t2: float) -> float:
        return self.state_voltage + (self.state_current + i) / (2 * self.value) * (t2 - t1)

    def get_jacobian(self, i: float, t1: float, t2: float) -> float:
        return (t2 - t1) / (2 * self.value)

    def update(self, v_i: np.ndarray, t1: float, t2: float):
        i = v_i[0]
        self.state_voltage = self.get_voltage(i, t1, t2)
        self.state_current = i

    @property
    def state(self):
        return self.state_voltage, self.state_current

    @state.setter
    def state(self, val):
        self.state_voltage, self.state_current = val


C = Capacitance



from dataclasses import dataclass, field

import numpy as np

from .TwoTerminalCurrentComponent import TwoTerminalCurrentComponent


@dataclass(eq=False)
class Inductance(TwoTerminalCurrentComponent):
    """
    Describes an ideal inductance (L*dI/dt = V)

    value:
        The inductance value in Henry.
    state_current:
        The momentary current. This is a state variable updated consecutively
        on each simulation iteration. However, you might set it ahead simulation
        to describe the inductance's initial state.
    state_voltage:
        The momentary voltage. Since this component uses the midpoint-method for
        calculation, this additional state variable is introduced for better accuracy.
        This is a state variable updated consecutively on each simulation iteration.
        However, you might set it ahead simulation to describe the inductance's initial state.
    """
    value: float

    state_current: float = field(default=0.0, repr=False)
    state_voltage: float = field(default=0.0, repr=False)

    def get_current(self, v: float, t1: float, t2: float) -> float:
        return self.state_current + (self.state_voltage + v) / (2 * self.value) * (t2 - t1)

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        return (t2 - t1) / (2 * self.value)

    def update(self, v_i: np.ndarray, t1: float, t2: float):
        v = v_i[0]
        self.state_current = self.get_current(v, t1, t2)
        self.state_voltage = v

    @property
    def state(self):
        return self.state_current, self.state_voltage

    @state.setter
    def state(self, val):
        self.state_current, self.state_voltage = val


L = Inductance



from collections.abc import MutableMapping

from ordered_set import OrderedSet


class OneToManyInvertibleMapping(MutableMapping):
    """
    A one-to-many mapping structure maintaining insertion order for values with fast reverse lookup.
    """

    class _ReverseDict(MutableMapping):
        def __init__(self, obverse):
            self._obverse = obverse
            self._reverse = dict()

        @property
        def obverse(self):
            return self._obverse

        def keys(self):
            return self._reverse.keys()

        def values(self):
            return self._reverse.values()

        def items(self):
            return self._reverse.items()

        def __getitem__(self, key):
            return self._reverse[key]

        def __setitem__(self, key, value):
            if key in self._reverse:
                if self._reverse[key] is value:
                    return

                raise ValueError('value is already obversely mapped to a different key - '
                                 'if you want to reset the relation, first delete key from value in the obverse '
                                 'mapping')

            self._reverse[key] = value

            if value in self._obverse._obverse:
                obverse_set = self._obverse[value]
            else:
                obverse_set = OrderedSet()
                self._obverse._obverse[value] = obverse_set

            obverse_set.add(key)

        def __delitem__(self, key):
            value = self._reverse[key]
            del self._reverse[key]

            obverse_set = self._obverse._obverse[value]
            obverse_set.remove(key)
            if not obverse_set:
                del self._obverse._obverse[value]

        def __iter__(self):
            return iter(self._reverse)

        def __len__(self):
            return len(self._reverse)

        def __repr__(self):
            return repr(self._reverse)

    def __init__(self, initialdata=None):
        self._obverse = dict()
        self._reverse = self._ReverseDict(self)

        if initialdata is not None:
            if isinstance(initialdata, OneToManyInvertibleMapping):
                for k, vs in initialdata.items():
                    for v in vs:
                        self.__setitem__(k, v)

            elif isinstance(initialdata, dict):
                for k, v in initialdata.items():
                    self.__setitem__(k, v)
            else:
                raise TypeError

    @property
    def reverse(self):
        return self._reverse

    def keys(self):
        return self._obverse.keys()

    def values(self):
        return self._obverse.values()

    def items(self):
        return self._obverse.items()

    def __getitem__(self, key):
        return self._obverse[key].copy()

    def __setitem__(self, key, value):
        if key in self._obverse:
            obverse_set = self._obverse[key]
        else:
            obverse_set = OrderedSet()
            self._obverse[key] = obverse_set

        obverse_set.add(value)

        self._reverse._reverse[value] = key

    def __delitem__(self, key):
        obverse_set = self._obverse[key]
        del self._obverse[key]

        for reverse in obverse_set:
            del self._reverse._reverse[reverse]

    def __iter__(self):
        return iter(self._obverse)

    def __len__(self):
        return len(self._obverse)

    def __repr__(self):
        return repr(self._obverse)



from math import isclose

from . import flatten, take


def compact(list_of_lists):
    """
    Compacts a list of lists.

    This is the same as flattening but also returns the compactification scheme, i.e. a list of lengths of the original
    sub-lists that were flattened. This is intended to be used with `uncompact`.
    """
    compacted = list(flatten(list_of_lists))
    compactification = list(map(len, list_of_lists))

    return compacted, compactification


def uncompact(compacted, compactification):
    """
    Reverses the compactification from `compact`.
    """
    fli = iter(compacted)
    return list(list(take(c, fli)) for c in compactification)


def intersperse(lists, elem):
    """
    Returns `lists[0], elem, lists[1], elem, lists[2], ..., elem, list[-1]`.
    """
    if len(lists) == 0:
        return

    listiter = iter(lists)
    yield from next(listiter)
    for l in lists:
        yield elem
        yield from l


def deresolve(x):
    """
    Removes redundant points that are too close to each other using `math.isclose`.

    :param x:
        Iterable to remove close points of.
    :return:
        A list of points with redundant close points removed.
    """

    xiter = iter(x)
    try:
        result = [next(xiter)]
    except StopIteration:
        return []

    for p in xiter:
        if not isclose(result[-1], p):
            result.append(p)

    return result



from typing import Callable, Optional

import numpy as np
from scipy.optimize import OptimizeResult


def solve(f: Callable,
          x0: np.ndarray,
          tol: float = 1.49012e-08,
          args=tuple(),
          stepbound: float = None,
          maxiter: Optional[int] = None,
          overflow_stepfactor: float = 0.5) -> OptimizeResult:

    if maxiter is None:
        maxiter = 200 * len(x0)

    x = x0.astype('float64')

    delta = 0
    delta_norm = 0

    for _ in range(maxiter):
        while True:
            try:
                F, J = f(x, *args)
            except OverflowError:
                if delta_norm < tol:
                    raise
                delta_norm *= overflow_stepfactor
                delta *= overflow_stepfactor
                x -= delta
            else:
                break

        delta = np.linalg.solve(J, -F)

        if any(np.isnan(d) for d in delta):
            return OptimizeResult({
                'x': x,
                'success': False,
                'fun': F,
                'fjac': J,
                'message': f'Encountered invalid float.',
            })

        delta_norm = np.linalg.norm(delta)

        if stepbound is not None:
            if delta_norm > stepbound:
                delta *= stepbound / delta_norm

        x += delta
        if delta_norm < tol:
            break
    else:
        F, J = f(x, *args)
        return OptimizeResult({
            'x': x,
            'success': False,
            'fun': F,
            'fjac': J,
            'message': f'Exceeded number of iterations ({maxiter}).',
        })

    F, J = f(x, *args)
    return OptimizeResult({
        'x': x,
        'success': True,
        'fun': F,
        'fjac': J,
        'message': 'The algorithm converged.',
    })



from typing import Hashable, Tuple, Dict, Container

import networkx as nx


class UniqueEdgeMultiDiGraph(nx.MultiDiGraph):
    """
    Extends the NetworkX MultiDiGraph with useful functions when edge-keys must be unique across all edges.
    """
    def __init__(self, data=None, **attr):
        self._edge_keys_to_nodes: Dict[Hashable, Tuple[Hashable, Hashable]] = {}
        super().__init__(data, **attr)

    def add_edge(self, u_for_edge: Hashable, v_for_edge: Hashable, key: Hashable, **attr) -> Hashable:
        if key in self._edge_keys_to_nodes:
            raise KeyError(f'Edge {key} already existing!')

        super().add_edge(u_for_edge, v_for_edge, key, **attr)

        self._edge_keys_to_nodes[key] = (u_for_edge, v_for_edge)

        return key

    def remove_edge(self, key: Hashable):
        if key not in self._edge_keys_to_nodes:
            raise nx.NetworkXError(f'Edge {key} does not exist.')

        super().remove_edge(*self._edge_keys_to_nodes[key], key)

        del self._edge_keys_to_nodes[key]

    def remove_edges_from(self, ebunch):
        try:
            for e in ebunch:
                self.remove_edge(e)
        except nx.NetworkXError:
            pass

    def merge_nodes(self, nodes: Container[Hashable], newnode: Hashable):
        """
        Merge nodes together into a new node.

        :param nodes:
            The nodes to merge together.
        :param newnode:
            A new node that replaces all nodes to merge unifying all their connections into it.
        """

        if newnode in nodes:
            raise nx.NetworkXError('new node must be distinct from nodes to merge')

        for _, target, key, data in list(self.out_edges(nodes, keys=True, data=True)):
            self.remove_edge(key)
            self.add_edge(newnode, target, key, **data)
        for source, _, key, data in list(self.in_edges(nodes, keys=True, data=True)):
            self.remove_edge(key)
            self.add_edge(source, newnode, key, **data)

        self.remove_nodes_from(nodes)

    def get_nodes(self, edge: Hashable) -> Tuple[Hashable, Hashable]:
        return self._edge_keys_to_nodes[edge]



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='respice',
    version='0.3.8',
    author='Mischa Krüger',
    author_email="makmanx64@gmail.com",
    description='Flexible and easy to use non-linear transient electric circuit simulator.',
    keywords='electronics circuit simulation non-linear transient steady-state time-domain',
    url='https://gitlab.com/Makman2/respice',
    project_urls={
        'Bug Tracker': 'https://gitlab.com/Makman2/respice/issues',
        'Source Code': 'https://gitlab.com/Makman2/respice/-/tree/master',
    },
    platforms='any',
    license='MIT',
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Topic :: Scientific/Engineering :: Physics",
    ],

    packages=find_packages(),
    python_requires='>=3.7',
    install_requires=read('requirements.txt').splitlines(),
    tests_require=read('test-requirements.txt').splitlines(),
    setup_requires=read('setup-requirements.txt').splitlines(),
    extras_require={
        'docs': read('docs-requirements.txt').splitlines(),
        'interactive': read('interactive-requirements.txt').splitlines(),
    }
)



import numpy as np

from .TransientEFMSimulation import TransientEFMSimulation


class TransientSteadyStateEFMSimulation(TransientEFMSimulation):
    def get_steadystate(self, component):
        """
        Returns the computed steady-state solution for a component.

        :param component:
            The component.
        :return:
            The steady-state vector.
        """
        return np.array(self._component_states[component][0][0])



from respice.analysis import Circuit
from respice.components import Resistance, Capacitance


class RC(Circuit):
    """
    Example RC circuit (without current supply).
    """

    def __init__(self, r: float, c: float, initial_v: float = 1.0):
        """
        :param r:
            Resistance value (measured in Ohm).
        :param c:
            Capacitance value (measured in Farad).
        :param initial_v:
            The initial voltage of the capacitance (measured in Volts).
        """
        super().__init__()

        self.R = Resistance(r)
        self.R.name = 'R'
        self.C = Capacitance(c, initial_v)
        self.C.name = 'C'

        self.add(self.R, 0, 1)
        self.add(self.C, 1, 0)



from dataclasses import dataclass
from math import pi
from typing import Optional

from .PeriodicSwitch import PeriodicSwitch


@dataclass(eq=False)
class SwitchRectangular(PeriodicSwitch):
    """
    A switch that switches on and off according to a rectangular-signal pattern.

    duty:
        The duty cycle. Meaningful values lie between 0 and 1.
    """
    duty: float = 0.5

    def get_switch_signal(self, phase: float) -> bool:
        return phase < 2 * pi * self.duty

    def next_switch_event(self, phase1: float, phase2: float) -> Optional[float]:
        if phase1 == 0.0:
            return 0.0

        dutyphase = self.duty * 2 * pi
        if phase1 <= dutyphase <= phase2:
            return dutyphase

        return None



import numpy as np

from respice.math.linalg import linsys_jac


def trap(x1, h, A1, b1, A2, b2):
    r"""
    Returns the next numerical integration step for a linear differential equation system.

    For linear systems, although the trapezoidal rule is an implicit integrator, it can be solved exactly.

    .. math::

        f(t, \boldsymbol{x}) = \dot{\boldsymbol{x}} = A(t) \boldsymbol{x} + \boldsymbol{b}(t) \\
        \boldsymbol{x}_2 = \boldsymbol{x}_1 + \frac12 h (f(t_1, \boldsymbol{x}_1) + f(t_2, \boldsymbol{x}_2))

    The solution to the trapezoidal formula becomes:

    .. math::

        \left( I - \frac12 h A(t_2) \right) \boldsymbol{x}_2 =
        \boldsymbol{x}_1 + \frac12 h (A(t_1) \boldsymbol{x}_1 + \boldsymbol{b}_1 + \boldsymbol{b}_2)

    And is simply solved as a linear equation.

    :param x1:
        State :math:`x(t)` at :math:`t = t_1`.
    :param h:
        Step size.
    :param A1:
        Linear coefficient matrix :math:`A(t)` for linear system at :math:`t = t_1`.
    :param b1:
        Ordinate vector :math:`\boldsymbol{b}(t)` for linear system at :math:`t = t_1`.
    :param A2:
        Linear coefficient matrix :math:`A(t)` for linear system at :math:`t = t_2`.
    :param b2:
        Ordinate vector :math:`\boldsymbol{b}(t)` for linear system at :math:`t = t_2`.
    :return:
        The next step.
    """
    tb = x1 + 0.5 * h * (A1 @ x1 + b1 + b2)
    tA = np.identity(len(tb)) - 0.5 * h * A2
    return np.linalg.solve(tA, tb)


def trap_jac(x2, h, A2, dA2, db2):
    r"""
    Returns the Jacobian matrix of the trapezoidal rule for a linear differential equation system for any given
    variable.

    Due to the trapezoidal rule being an implicit method, the Jacobian matrix becomes a quite complex expression.
    Since the states at :math:`t = t_1` are considered to be constant (they represent an already calculated step),
    only states at :math:`t = t_2` are relevant.

    .. math::

        \left(
            I - \frac12 h A(t_2)
        \right)
        J_{\boldsymbol{x}_2}(\boldsymbol{y}) =
        \frac12 h
        \left(
            \left. \frac{\partial \boldsymbol{b}}{\partial \boldsymbol{y}} \right|_{t_2} +
            \left(
                \left. \frac{\partial A}{\partial \boldsymbol{y}} \right|_{t_2} \boldsymbol{x_2}
            \right )^{T}
        \right)

    :param x2:
        Next state :math:`x(t)` at :math:`t = t_2`.
        This is the solution of `trap(x1, h, A1, b1, A2, b2)`.
    :param h:
        Step size.
    :param A2:
        Linear coefficient matrix :math:`A(t)` for linear system at :math:`t = t_2`.
    :param dA2:
        A 3rd-order tensor containing the element-wise derivatives of each matrix entry in :math:`A(y)`
        over each :math:`y`.
    :param db2:
        The derivative matrix of :math:`b(y)`.
    :return:
        The Jacobian.
    """
    # The (0.5 * h) part can be pulled out of the linear Jacobian expression according to the formulas.
    # This improves performance.
    return 0.5 * h * linsys_jac(x2,
                                np.identity(len(x2)) - 0.5 * h * A2,
                                -dA2,
                                db2)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Hashable, Tuple, Sequence

import numpy as np

from .Branch import Branch, CurrentBranch
from .Component import Component


@dataclass(eq=False)
class TwoTerminalCurrentComponent(Component):
    """
    The base class for all two-terminal electrical components that have a voltage-to-current characteristic.

    This class avoids you to maintain `Branch`s for this simple two-terminal case. Instead of creating a branch,
    you can immediately override methods like `get_current` or `update` and access voltages and currents via this
    component itself (`component.v` and `component.i`).
    """
    _branch: _TwoTerminalBranch = field(init=False, repr=False)

    @dataclass(eq=False)
    class _TwoTerminalBranch(CurrentBranch):
        """
        The branch that represents the two-terminal component.

        Effectively mirrors all functions like `get_current` back to the parent component.
        """
        def get_current(self, v_i: np.ndarray, t1: float, t2: float) -> float:
            return self.component.get_current(v_i[0], t1, t2)

        def get_jacobian(self, v_i: np.ndarray, t1: float, t2: float) -> Sequence[float]:
            return [self.component.get_jacobian(v_i[0], t1, t2)]

        def __str__(self):
            return f'{str(self.component)}'

    def __post_init__(self):
        self._branch = self._TwoTerminalBranch(self)

    def connect(self, terminal1: Hashable, terminal2: Hashable) -> List[Tuple[Hashable, Hashable, Branch]]:
        return [(terminal1, terminal2, self._branch)]

    def get_current(self, v: float, t1: float, t2: float) -> float:
        """
        Calculated current depending on given voltage and time step.

        Shall return the current at present point in time `t2`.

        :param v:
            Voltage (measured in Volts).
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The current (measured in Ampere).
        """
        raise NotImplementedError

    def get_jacobian(self, v: float, t1: float, t2: float) -> float:
        """
        Returns the derivative (a.k.a. 1D Jacobian) of `get_voltage` over the voltage `v`.

        For the numerical algorithm to quickly converge with more accuracy, it is necessary for electrical circuit
        problems to supply the derivations of each current / voltage function.

        See also https://en.wikipedia.org/wiki/Derivative and
        https://en.wikipedia.org/wiki/Jacobian_matrix_and_determinant

        :param v:
            Voltage (measured in Volts).
        :param t1:
            Previous point in time (measured in seconds).
        :param t2:
            Present point in time (measured in seconds).
        :return:
            The derivative value (measured in Ampere per second).
        """
        raise NotImplementedError

    @property
    def default_branch(self):
        """
        Returns the one and only branch of a two terminal component.
        """
        return self._branch
